package org.bolson.vote;
/**
 * Top level abstract class for ranked or rated voting systems.
 */

public abstract class VotingSystem implements java.io.Serializable {
    /** number of candidates */
    protected int numc;

    /** In addition to interfaced methods, there shall be a constructor taking 
     * the integer number N of choices to be voted on. */
    public VotingSystem( int numCandidates ) {
	numc = numCandidates;
    };

    /**
     * @param rating An array int[numc].
     *	rating[i] is the rating of choice i.
     *  A rating can be any int value. Higher means more preferred.
     *  NO_VOTE for unrated.
     * @return 0 on success
     */
    public abstract int voteRating( int rating[] );
    /**
     * @param rating An array int[N].
     *	rating[i] is the rating of choice i.
     *  A rating can be any int value. Higher means more preferred.
     *  NO_VOTE for unrated.
     * @return 0 on success
     */
    public abstract int voteRating( float rating[] );
    /**
     * @param rating An array int[N].
     *	rating[i] is the rating of choice i.
     *  A rating can be any int value. Higher means more preferred.
     *  NO_VOTE for unrated.
     * @return 0 on success
     */
    public abstract int voteRating( double rating[] );
    /**
     * @param ranking An array int[N].
     *	ranking[i] is the ranking of choice i.
     *  rankings are 1 (most preferred) through N (least).
     *  INT_MAX, N, or NO_VOTE for unspecified.
     * @return 0 on success
     */
    public abstract int voteRanking( int ranking[] );

    /**
     * squash (sum of absolute value of vote) to (max - min)
     * @param rating vote to check
     * @param min minimum (inclusive) value allowed for rating
     * @param max maximum (inclusive) value allowed for rating
     * @return true if rating.length == numc AND all values in rating are >= min and <= max
     */
    public boolean normalizeRatedVote( int rating[], int min, int max ) {
	if ( rating.length != numc ) return false;
	for ( int i = 0; i < rating.length; i++ ) {
	    if ( rating[i] == NO_VOTE ) {
		// OK
	    } else if ( rating[i] < min ) {
		return false;
	    } else if ( rating[i] > max ) {
		return false;
	    } // else OK
	}
	// didn't find anything bad, must be good
	return true;
    }
    /**
     * Checks that rating is valid.
     * @param rating vote to check
     * @param min minimum (inclusive) value allowed for rating
     * @param max maximum (inclusive) value allowed for rating
     * @return true if rating.length == numc AND all values in rating are >= min and <= max
     */
    public boolean isValidRatedVote( int rating[], int min, int max ) {
	if ( rating.length != numc ) return false;
	for ( int i = 0; i < rating.length; i++ ) {
	    if ( rating[i] == NO_VOTE ) {
		// OK
	    } else if ( rating[i] < min ) {
		return false;
	    } else if ( rating[i] > max ) {
		return false;
	    } // else OK
	}
	// didn't find anything bad, must be good
	return true;
    }
    /**
     * Checks that ranking is valid.
     * @param rating vote to check
     * @return true if rating.length == numc AND all values in rating are >= 1 and <= numc
     */
    public boolean isValidRankedVote( int ranking[] ) {
	if ( ranking.length != numc ) return false;
	for ( int i = 0; i < ranking.length; i++ ) {
	    if ( ranking[i] == NO_VOTE ) {
		// OK
	    } else if ( ranking[i] < 1 ) {
		return false;
	    } else if ( ranking[i] > numc ) {
		return false;
	    } // else OK
	}
	// didn't find anything bad, must be good
	return true;
    }
    
    /**
     * @return indecies of winners (hopefully 1 of them)
     */
    public abstract int[] getWinners();

    /**
     * A more interesting representation.
     * Include the names of the choices voted upon in the representation.
     * @param names The names of the choices.
     * @return state, with names!
     */
    public abstract String toString( String names[] );

    /**
     * A fancy html representation.
     * Include the names of the choices voted upon in the representation.
     * @param names The names of the choices. May be null.
     * @return state, with names!
     */
    public abstract String htmlSummary( String names[] );

    /**
     * This value represents that the voter made no action on some choice.
     * This is the default value. On a YES/NO or HOW MUCH the answer was no answer.
     */
    public static final int NO_VOTE = 0x80000000;
}
