/* This file is placed in the public domain without warrantee by it's creator, Brian Olson. */
#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <time.h>
#include <unistd.h>
#include <math.h>
#include <string.h>
#include <errno.h>

#include "Condorcet.h"
#include "BordaVotingSystem.h"
#include "RawRating.h"
#include "IRV.h"
#include "Histogram.h"

char** readNames( char* filename );

class runParams {
public:
    bool rating;
    bool isFloat;
    bool isBin;
    bool scriptMode;
    bool htmlOut;
    bool printState;
    bool quiet;
    VotingSystem* (*fac)(int);
    FILE* fin;
    int numc;
    int numWinners;
    int* p;
    bool histogram;
    int histBuckets;
    char** names;
    int histyle;
    
    runParams();
};

runParams::runParams()
    : rating( false ), isFloat( false ), isBin( false ), scriptMode( false ),
      htmlOut( false ), printState( false ), quiet( false ),
      fac( BordaVotingSystem::newBordaVotingSystem ), fin( NULL ), numc( -1 ),
      numWinners( -1 ), p( NULL ), histogram( false ), histBuckets( -1 ),
      names( NULL ), histyle( 2 )
{}

int voteFromTextFile( runParams* r ) {
    VotingSystem* tab;
    Histogram* hist = NULL;
    int* p;
    int numc = r->numc;
    if ( !r->fin ) return -1;
    if ( r->isBin ) {
	if ( numc == -1 ) fread( &numc, sizeof(numc), 1, r->fin );
    } else {
	if ( numc == -1 ) fscanf( r->fin, "%d", &numc );
    }
    tab = r->fac( numc );
    if ( r->histogram ) {
	int nbuck = r->histBuckets;
	double min, max;
	if ( nbuck == -1 ) {
	    if ( r->rating ) {
		nbuck = 11;
	    } else {
		nbuck = numc;
	    }
	}
	if ( r->rating ) {
	    if ( r->isFloat ) {
		min = -1.0;
		max = 1.0;
	    } else {
		min = 0;
		max = 10;
	    }
	} else {
	    min = 1;
	    max = numc;
	}
	hist = new Histogram( numc, nbuck, min, max );
    }
    p = new int[numc];
    while ( 1 ) {
	if ( r->isBin ) {
	    int err;
	    for ( int i = 0; i < numc; i++ ) {
		if ( r->isFloat ) {
		    float tf;
		    err = fread( &tf, sizeof(float), 1, r->fin );
		    p[i] = (int)(tf * INT_MAX);
		} else {
		    err = fread( &(p[i]), sizeof(int), 1, r->fin );
		}
		if ( err <= 0 ) goto noMoreVotes;
	    }
	} else {
	    char* line;
	    char* pos;
	    char* npos;
	    size_t llen;
	    line = fgetln( r->fin, &llen );
	    if ( line == NULL ) {
		goto noMoreVotes;
	    }
	    pos = line;
	    for ( int i = 0; i < numc; i++ ) {
		if ( *pos == '\t' || *pos == '\n' ) {
		    p[i] = NO_VOTE;
		    pos++;
		} else if ( r->isFloat ) {
		    double td;
		    td = strtod( pos, &npos );
		    if ( pos == npos || errno == ERANGE ) {
			goto noMoreVotes;
		    }
		    pos = npos;
		    if ( *pos == '\t' ) pos++;
		    p[i] = (int)(td * INT_MAX);
		} else {
		    long tl;
		    tl = strtol( pos, &npos, 10 );
		    if ( pos == npos || errno == ERANGE ) {
			goto noMoreVotes;
		    }
		    pos = npos;
		    if ( *pos == '\t' ) pos++;
		    p[i] = tl;
		}
		if ( pos - line > llen ) {
		    if ( i != numc - 1 ) {
			fprintf( stderr, "warning, short line\n" );
		    }
		    break;
		}
	    }
	}
#if 0
	fprintf( stderr, "voting %s\t", r->rating ? "rating " : "ranking" );
	for ( int i = 0; i < numc; i++ ) {
	    if ( p[i] == NO_VOTE ) {
		fprintf( stderr, "\t");
	    } else {
		fprintf( stderr, "%d\t", p[i] );
	    }
	}
	fprintf( stderr, "\n");
#endif
	if ( r->rating ) {
	    tab->voteRating( p );
	} else {
	    tab->voteRanking( p );
	}
	if ( hist ) {
	    if ( r->rating ) {
		hist->addRating( p );
	    } else {
		hist->addRanking( p );
	    }
	}
    }
  noMoreVotes:
    fclose( r->fin );
    r->fin = NULL;
    r->numWinners = tab->getWinners( p );
    r->p = p;
    if ( r->printState ) {
	tab->print( stdout, r->names );
    }
    if ( r->htmlOut ) {
	tab->htmlSummary( stdout, r->names );
    }
    if ( hist ) {
	hist->print( stdout, r->histyle, r->names );
    }
    delete tab;
    return 0;
}

void printWinners( runParams* r ) {
    if ( r->scriptMode ) {
	for ( int i = 0; i < r->numWinners; i++ ) {
	    printf("%d\n", r->p[i] );
	}
    } else if ( r->numWinners == 1 ) {
	if ( r->names ) {
	    printf("Winner: %s\n", r->names[r->p[0]] );
	} else {
	    printf("winner, choice number %d (of 1..%d)\n", r->p[0] + 1, r->numc );
	}
    } else {
	if ( r->names ) {
	    printf("Winners: %s", r->names[r->p[0]] );
	    for ( int i = 1; i < r->numWinners; i++ ) {
		printf(", %s", r->names[r->p[0]] );
	    }
	} else {
	    printf("%d winners, numbers %d", r->numWinners, r->p[0] + 1 );
	    for ( int i = 1; i < r->numWinners; i++ ) {
		printf(", %d", r->p[i] + 1 );
	    }
	}
	printf(" (of 1..%d)\n", r->numc );
    }
    delete [] r->p;
    r->p = NULL;
}

char* optstring = "bBcdDfhHi:In:N:pqRrsST:w";
const char* const usageText = "vote [-bBcdfhHpqRrw][-n number of choices][-i vote file][vote file ...]\n";
const char* const helpText =
"\t-b\tbinary\n"
"\t-B\ttext (default)\n"
"\t-c\tCondorcet\n"
"\t-D\tBorda count (default)\n"
"\t-d\tinteger (default)\n"
"\t-f\tfloat (-1.0 .. 1.0)\n"
"\t-h\tthis help\n"
"\t-H\toutput html table\n"
"\t-i f\timmediately load and run vote file f\n"
"\t-I\tInstant Runoff Vote\n"
"\t-n N\tnumber of choices\n"
"\t-N f\tfile with names of choices\n"
"\t-p\tprint full state, not just winner\n"
"\t-q\tquiet, don't print winner\n"
"\t-R\tnumbers are rankings (1 best) (default)\n"
"\t-r\tnumbers are ratings (higher better)\n"
"\t-s\tscript friendly output (zero indexed \\n separated list of winners)\n"
"\t-S\tmake histogram of ranked or rated votes\n"
"\t-T #\thistogram style #\n"
"\t-w\tRaw Sum of ratings\n"
;

FILE* myOpen( const char* optarg ) {
    if ( !strcmp(optarg,"-") ) {
	return stdin;
    } else {
	FILE* fin = fopen( optarg, "r" );
	if ( !fin ) {
	    perror( optarg );
	}
	return fin;
    }
}


int main( int argc, char** argv ) {
    int c;
    extern char* optarg;
    extern int optind;
    
    runParams r;
    
    while ( (c = getopt(argc,argv,optstring)) != -1 ) switch ( c ) {
    case 'B':
	r.isBin = false;
	break;
    case 'b':
	r.isBin = true;
	break;
    case 'c':
	r.fac = Condorcet::newCondorcet;
	break;
    case 'D':
	r.fac = BordaVotingSystem::newBordaVotingSystem;
	break;
    case 'd':
	r.isFloat = false;
	break;
    case 'f':
	r.isFloat = true;
	break;
    case 'h':	// help
	printf(usageText);
	printf(helpText);
	break;
    case 'H':
	r.htmlOut = true;
	break;
    case 'i':
	r.fin = myOpen( optarg );
    	r.numWinners = voteFromTextFile( &r );
	printWinners( &r );
	break;
    case 'I':
	r.fac = IRV::newIRV;
	break;
    case 'n':
	r.numc = atoi( optarg );
	break;
    case 'N':
	// read names
	r.names = readNames( optarg );
	break;
    case 'p':
	r.printState = true;
	break;
    case 'q':
	r.quiet = true;
	break;
    case 'R':
	r.rating = false;
	break;
    case 'r':
	r.rating = true;
	break;
    case 's':
	r.scriptMode = true;
	break;
    case 'S':
	r.histogram = true;
	break;
    case 'T':
	r.histyle = atoi( optarg );
	r.histogram = true;
	break;
    case 'w':
	r.fac = RawRating::newRawRating;
	break;
    default:
	printf("unknown option \'%c\'\n", c );
	return 1;
    }

    for ( ; optind < argc; optind++ ) {
	optarg = argv[optind];
	r.fin = myOpen( optarg );
    	voteFromTextFile( &r );
	if ( ! r.quiet ) {
	    printWinners( &r );
	}
    }

    return 0;
}

char** readNames( char* filename ) {
    char** toret;
    int namecap = 100;
    int nami = 0;
    char* cur;
    size_t curLen;
    FILE* fin;

    fin = fopen( filename, "r" );
    if ( fin == NULL ) {
	perror( filename );
	return NULL;
    }
    
    toret = (char**)malloc( sizeof(char*) * namecap );
    while ( (cur = fgetln( fin, &curLen )) != NULL ) {
	if ( cur[curLen-1] == '\n' ) {
	    curLen--;
	}
	toret[nami] = (char*)malloc( curLen + 1 );
	memcpy( toret[nami], cur, curLen );
	toret[nami][curLen] = '\0';
	nami++;
	if ( nami == namecap ) {
	    namecap *= 2;
	    toret = (char**)reallocf( toret, sizeof(char*) * namecap );
	    if ( toret == NULL ) {
		perror("realloc");
		return NULL;
	    }
	}
    }
    fclose( fin );
    toret = (char**)realloc( toret, sizeof(char*) * nami );
    return toret;
}

