#include "WorkQueue.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

WorkUnit* WorkSource::newWorkUnit() {
    return newWorkUnit( 0 );
}
WorkUnit::WorkUnit( int nv, int nc, int ns, float e, int t )
    : numv( nv ), numc( nc ), seats( ns ), error( e ), trials( t ), next( (WorkUnit*)0 )
{}
void WorkUnit::print( FILE* f ) {
    fprintf( f, "%d voters, %d choices, %d seats, %0.2f error, %d trials\n", numv, numc, seats, error, trials );
}
void WorkUnit::printSummaryHeader( FILE* f, int totalTrials, int style ) {
    if ( style == 4 ) {
        fprintf( f, "%d candidates, %d seats, %d voters, %d trial elections, +/- %f confusion error<br><table><tr><th>Name</th><th>Avg. Happiness<br>(higher better)</th><th>Reliability Std.<br>(lower better)</th><th>Consensus Std.<br>(lower better)</th></tr>\n", numc, seats, numv, trials, error );
    } else if ( style != 0 ) {
	if ( error < 0 ) {
	    fprintf( f, "%d candidates, %d seats, %d voters, %d trial elections\n",
	      numc, seats, numv, trials );
	} else {
	    fprintf( f, "%d candidates, %d seats, %d voters, %d trial elections, +/- %f confusion error\n",
	      numc, seats, numv, trials, error );
	}
    }
}
void WorkUnit::printSummaryFooter( FILE* f, int totalTrials, int style ) {
    if ( style == 4 ) {
        fprintf( f, "<tr><th>Name</th><th>Avg. Happiness<br>(higher better)</th><th>Reliability Std.<br>(lower better)</th><th>Consensus Std.<br>(lower better)</th></tr></table>\n" );
    }
}

WorkQueue::WorkQueue()
    : head( NULL ), tail( NULL )
{}
WorkQueue::~WorkQueue() {
    WorkUnit* cur = head;
    while ( cur ) {
	WorkUnit* t;
	t = cur->next;
	delete cur;
	cur = t;
    }
}

void WorkQueue::add( WorkUnit* it ) {
    it->next = NULL;
    if ( tail == NULL ) {
	head = tail = it;
    } else {
	tail->next = it;
	tail = it;
    }
}

WorkUnit* WorkQueue::pop() {
    if ( head == NULL ) {
	return NULL;
    } else {
	WorkUnit* toret = head;
	head = head->next;
	if ( head == NULL ) tail = NULL;
	return toret;
    }
}

WorkUnit* WorkQueue::newWorkUnit( WorkUnit* wuin ) {
    if ( wuin != NULL ) {
	delete wuin;
    }
    return pop();
}

#ifndef DIM
#define DIM(n) (sizeof(n)/sizeof(n[0]))
#endif

Steps::Steps()
    : v( 0 ), c( 0 ), e( -1 )
{
#if DEVEL
    static const int dvsteps[] = { 100 };
    static const int dcsteps[] = { 5 };
    static const int dssteps[] = { 2 };
    static const float desteps[] = { 0.1 };
#else
    static const int dvsteps[] = { 100, 1000 };
    static const int dcsteps[] = { 5, 6, 7, 8 };
    static const int dssteps[] = { 2, 3, 4 };
    static const float desteps[] = { -1, 0.00, 0.01, 0.1, 0.25, 0.5, 1.0, 1.5, 2.0 };
#endif
    vsteps = (int*)malloc( sizeof(dvsteps) );
    memcpy( vsteps, dvsteps, sizeof(dvsteps) );
    vstepsLen = DIM(dvsteps);
    csteps = (int*)malloc( sizeof(dcsteps) );
    memcpy( csteps, dcsteps, sizeof(dcsteps) );
    cstepsLen = DIM(dcsteps);
    seatsteps = (int*)malloc( sizeof(dssteps) );
    memcpy( seatsteps, dssteps, sizeof(dssteps) );
    seatstepsLen = DIM(dssteps);
    esteps = (float*)malloc( sizeof(desteps) );
    memcpy( esteps, desteps, sizeof(desteps) );
    estepsLen = DIM(desteps);
    n = 10000;
}
Steps::~Steps() {
    free( vsteps );
    free( csteps );
    free( seatsteps );
    free( esteps );
}

#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

Steps::Steps( const char* vlist, const char* clist, const char* seatlist, const char* elist, int nIn )
    : vstepsLen( 1 ), cstepsLen( 1 ), seatstepsLen( 1 ), estepsLen( 1 ),
      v( 0 ), c( 0 ), s( 0 ), e( -1 ), n( nIn )
{
    if ( vlist ) {
	parseV( vlist );
    } else {
	vsteps = (int*)malloc( sizeof(int) );
	vstepsLen = 0;
    }
    if ( clist ) {
	parseC( clist );
    } else {
	csteps = (int*)malloc( sizeof(int) );
	cstepsLen = 0;
    }
    if ( seatlist ) {
	parseSeats( seatlist );
    } else {
	seatsteps = (int*)malloc( sizeof(int) );
	seatstepsLen = 0;
    }
    if ( elist ) {
	parseE( elist );
    } else {
	esteps = (float*)malloc( sizeof(float) );
	estepsLen = 0;
    }
}
void Steps::parseV( const char* vlist ) {
    const char* cur;
    int pos;
    vstepsLen = 1;
    cur = vlist;
    while ( *cur != '\0' ) {
	if ( isspace( *cur )) {
	    vstepsLen++;
	    do {
		cur++;
	    } while ( isspace( *cur ));
	} else {
	    cur++;
	}
    }
    vsteps = (int*)malloc( sizeof(int) * vstepsLen );
    cur = vlist;
    pos = 0;
    ((int*)vsteps)[pos] = atoi( cur );
    while ( *cur != '\0' ) {
	if ( isspace( *cur )) {
	    pos++;
	    do {
		cur++;
	    } while ( isspace( *cur ));
	    ((int*)vsteps)[pos] = atoi( cur );
	}
	cur++;
    }
}

void Steps::parseC( const char* clist ) {
    const char* cur;
    int pos;
    cur = clist;
    cstepsLen = 1;
    while ( *cur != '\0' ) {
	if ( isspace( *cur )) {
	    cstepsLen++;
	}
	cur++;
    }
    csteps = (int*)malloc( sizeof(int) * cstepsLen );
    cur = clist;
    pos = 0;
    ((int*)csteps)[pos] = atoi( cur );
    while ( *cur != '\0' ) {
	if ( isspace( *cur )) {
	    pos++;
	    do {
		cur++;
	    } while ( isspace( *cur ));
	    ((int*)csteps)[pos] = atoi( cur );
	}
	cur++;
    }
}

void Steps::parseSeats( const char* clist ) {
    const char* cur;
    int pos;
    cur = clist;
    seatstepsLen = 1;
    while ( *cur != '\0' ) {
	if ( isspace( *cur )) {
	    seatstepsLen++;
	}
	cur++;
    }
    seatsteps = (int*)malloc( sizeof(int) * seatstepsLen );
    cur = clist;
    pos = 0;
    ((int*)seatsteps)[pos] = atoi( cur );
    while ( *cur != '\0' ) {
	if ( isspace( *cur )) {
	    pos++;
	    do {
		cur++;
	    } while ( isspace( *cur ));
	    ((int*)seatsteps)[pos] = atoi( cur );
	}
	cur++;
    }
}

void Steps::parseE( const char* elist ) {
    const char* cur;
    int pos;
    cur = elist;
    estepsLen = 1;
    while ( *cur != '\0' ) {
	if ( isspace( *cur )) {
//	    printf("espace '%c'\n", *cur );
	    estepsLen++;
	}
	cur++;
    }
    esteps = (float*)malloc( sizeof(int) * estepsLen );
    cur = elist;
    pos = 0;
    ((float*)esteps)[pos] = atof( cur );
    while ( *cur != '\0' ) {
	if ( isspace( *cur )) {
	    pos++;
	    do {
		cur++;
	    } while ( isspace( *cur ));
	    ((float*)esteps)[pos] = atof( cur );
	}
	cur++;
    }
}

inline int present( int v, int* arr, int len ) {
    for ( int i = 0; i < len; i++ ) {
	if ( arr[i] == v ) {
	    return 1;
	}
    }
    return 0;
}
inline int present( float v, float* arr, int len ) {
    for ( int i = 0; i < len; i++ ) {
	if ( arr[i] == v ) {
	    return 1;
	}
    }
    return 0;
}
inline void insert( int v, int* arr, int& len ) {
    int i = len - 1;
    len++;
    for ( ; i >= 0; i-- ) {
	if ( arr[i] > v ) {
	    arr[i+1] = arr[i];
	} else {
	    arr[i+1] = v;
	    return;
	}
    }
    arr[0] = v;
}
inline void insert( float v, float* arr, int& len ) {
    int i = len - 1;
    len++;
    for ( ; i >= 0; i-- ) {
	if ( arr[i] > v ) {
	    arr[i+1] = arr[i];
	} else {
	    arr[i+1] = v;
	    return;
	}
    }
    arr[0] = v;
}
void Steps::addIfNotPresent( int v, int c, int s, float e ) {
    if ( ! present( v, vsteps, vstepsLen ) ) {
	vsteps = (int*)realloc( vsteps, (vstepsLen + 1) * sizeof(int) );
	assert( vsteps );
	insert( v, vsteps, vstepsLen );
    }
    if ( ! present( c, csteps, cstepsLen ) ) {
	csteps = (int*)realloc( csteps, (cstepsLen + 1) * sizeof(int) );
	assert( csteps );
	insert( c, csteps, cstepsLen );
    }
    if ( ! present( e, esteps, estepsLen ) ) {
	esteps = (float*)realloc( esteps, (estepsLen + 1) * sizeof(float) );
	assert( esteps );
	insert( e, esteps, estepsLen );
    }
}

WorkUnit* Steps::newWorkUnit( WorkUnit* toret ) {
    e++;
    if ( e == estepsLen ) {
	e = 0;
	c++;
    }
    if ( c == cstepsLen ) {
	c = 0;
	v++;
    }
    if ( v == vstepsLen ) {
//	v = 0;
	return NULL;
    }
    if ( toret == NULL ) {
	toret = new WorkUnit( vsteps[v], csteps[c], seatsteps[s], esteps[e], n );
    } else {
	toret->set( vsteps[v], csteps[c], seatsteps[s], esteps[e], n );
    }
    return toret;
}

void Steps::print( void* vf ) {
    FILE* f = (FILE*)vf;
    int i;
    if ( vsteps == NULL || vstepsLen == 0 ) {
	fprintf( f, "no voter steps" );
    } else {
	fprintf( f, "voter steps: %d", vsteps[0] );
	for ( i = 1; i < vstepsLen; i++ ) {
	    fprintf( f, ", %d", vsteps[i] );
	}
    }
    if ( csteps == NULL || cstepsLen == 0 ) {
	fprintf( f, "\nno choice steps" );
    } else {
	fprintf( f, "\nchoice steps: %d", csteps[0] );
	for ( i = 1; i < cstepsLen; i++ ) {
	    fprintf( f, ", %d", csteps[i] );
	}
    }
    if ( seatsteps == NULL || seatstepsLen == 0 ) {
	fprintf( f, "\no seat steps" );
    } else {
	fprintf( f, "\nseat steps: %d", seatsteps[0] );
	for ( i = 1; i < seatstepsLen; i++ ) {
	    fprintf( f, ", %d", seatsteps[i] );
	}
    }
    if ( esteps == NULL || estepsLen == 0 ) {
	fprintf( f, "\nno error steps" );
    } else {
	fprintf( f, "\nerror steps: %f", esteps[0] );
	for ( i = 1; i < estepsLen; i++ ) {
	    fprintf( f, ", %f", esteps[i] );
	}
    }
    fprintf( f, "\npos: v %d, c %d, s %d, e %d, iter=%d\n", v, c, s, e, n );
    fprintf( f, "val: v %d, c %d, s %d, e %f, iter=%d\n", vsteps[v], csteps[c], seatsteps[s], esteps[e], n );
}
void Steps::print() {
    print( stdout );
}

#if STEPS_TEST_MAIN
int main( int argc, char** argv ) {
    Steps* s;
    s = new Steps( argv[1], argv[2], argv[3], 1000 );
    WorkUnit* wu;
    while ( (wu = s->newWorkUnit()) != NULL ) {
	wu->print( stdout );
	delete wu;
    }
}
#endif

ThreadSafeWorkSource::ThreadSafeWorkSource( WorkSource* source )
    : realSource( source )
{
    pthread_mutex_init( &m, NULL );
}
ThreadSafeWorkSource::~ThreadSafeWorkSource() {
    pthread_mutex_destroy( &m );
}

WorkUnit* ThreadSafeWorkSource::newWorkUnit( WorkUnit* toret ) {
    pthread_mutex_lock( &m );
    toret = realSource->newWorkUnit( toret );
    pthread_mutex_unlock( &m );
    return toret;
}
