/* This file is placed in the public domain without warrantee by it's creator, Brian Olson. */
#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <time.h>
#include <unistd.h>
#include <math.h>
#include <string.h>
#include <signal.h>
#include <pthread.h>

#include "Voter.h"
#include "VoterSim.h"
#include "VotingSystem.h"

#include "OneVotePickOne.h"
#include "RankedVotePickOne.h"
#include "AcceptanceVotePickOne.h"
#include "FuzzyVotePickOne.h"
#include "InstantRunoffVotePickOne.h"
#include "Condorcet.h"
#include "IRNR.h"
#include "RandomElection.h"

#include "ResultFile.h"
#include "DBResultFile.h"
#include "ThreadSafeDBRF.h"
#include "WorkQueue.h"

void Voter::setWithError( const Voter& it, float error ) {
    for ( int i = 0; i < preflen; i++ ) {
        preference[i] = it.preference[i] + ((random() / ((double)LONG_MAX)) - 0.5) * 2.0 * error;
        if ( preference[i] < -1.0 ) {
            preference[i] = -1.0;
        } else if ( preference[i] > 1.0 ) {
            preference[i] = 1.0;
        }
    }
}
void Voter::set( const Voter& it, Strategy* s ) {
    if ( s ) {
	s->apply( preference, it );
    } else {
	for ( int i = 0; i < it.preflen; i++ ) {
	    preference[i] = it.preference[i];
	}
    }
}

// "pretty" text
void Voter::print( void* fi ) const {
    FILE* f = (FILE*)fi;
    fprintf(f,"( %g", preference[0]);
    for ( int i = 1; i < preflen; i++ ) {
        fprintf(f,", %g",preference[i]);
    }
    fprintf(f," )");
}
// ugly, parse friendly text
void Voter::dump( void* fi ) const {
    FILE* f = (FILE*)fi;
    fprintf(f,"%.15g", preference[0]);
    for ( int i = 1; i < preflen; i++ ) {
        fprintf(f,"\t%.15g",preference[i]);
    }
    fprintf(f,"\n");
}
// text
void Voter::undump( void* fi, int len ) {
    FILE* f = (FILE*)fi;
    if ( len != preflen || preference == NULL ) {
	if ( preference ) {
	    delete [] preference;
	}
	preference = new float[len];
	preflen = len;
    }
    for ( int i = 0; i < len; i++ ) {
	fscanf( f, "%f",&(preference[i]));
    }
}
// binary
void Voter::write( void* fi ) const {
    FILE* f = (FILE*)fi;
    fwrite( preference, sizeof(*preference), preflen, f );
}

// binary
void Voter::read( void* fi, int len ) {
    FILE* f = (FILE*)fi;
    if ( len != preflen || preference == NULL ) {
	if ( preference ) {
	    delete [] preference;
	}
	preference = new float[len];
	preflen = len;
    }
    fread( preference, sizeof(*preference), len, f );
}

int Voter::getMax() const {
    int i;
    float m = preference[0];
    int toret = 0;
    
    for ( i = 1; i < preflen; i++ ) {
        if ( preference[i] > m ) {
            m = preference[i];
            toret = i;
        }
    }
    return toret;
}

Voter::Voter()
    : preference( NULL ), preflen( 0 ) {
};

Voter::Voter( int numCandidates )
    : preflen( numCandidates ) {
    preference = new float[preflen];
}
// -1.0 .. 1.0
double* prefBias;
inline double biasedRandom( double bias ) {
    // [-1.0 .. 1.0] ==> [Pi/2 .. 0]
    bias = (1.0 - bias) * (M_PI_4);
    // random in [0.0 .. 1.0] biased by exponentiation to same range, expanded to [-1.0 .. 1.0]
    return (pow( (random() / ((double)LONG_MAX)), tan( bias ) ) - 0.5) * 2.0;
}

void Voter::randomize() {
    for ( int i = 0; i < preflen; i++ ) {
#if 01
        preference[i] = ((random() / ((double)LONG_MAX)) - 0.5) * 2.0;
#elif 0
        
#else
        // scale [0.0 - 1.0) to [-1.0 - 1.0)
        preference[i] = (drand48() - 0.5) * 2.0;
#endif
    }
}

void Voter::randomizeWithFirstChoice( int first ) {
    int i;
    for ( i = 0; i < preflen; i++ ) {
#if 01
        preference[i] = ((random() / ((double)LONG_MAX)) - 0.5) * 2.0;
#elif 0
        
#else
        // scale [0.0 - 1.0) to [-1.0 - 1.0)
        preference[i] = (drand48() - 0.5) * 2.0;
#endif
    }
    // sort highest value into position of first choice
    for ( i = 0; i < preflen; i++ ) if ( i != first ) {
	if ( preference[i] > preference[first] ) {
	    float t = preference[i];
	    preference[i] = preference[first];
	    preference[first] = t;
	}
    }
}

Voter::~Voter() {
    if ( preference ) {
	delete preference;
    }
}

void Voter::setStorage( float* p, int pl ) {
    preference = p;
    preflen = pl;
}

void Voter::clearStorage() {
    preference = NULL;
    preflen = 0;
}


/* for float p[l]; return index of max */
#if 0
int getMax( float* p, int l ) {
    int i;
    float m = p[0];
    int toret = 0;
    
    for ( i = 1; i < l; i++ ) {
        if ( p[i] > m ) {
            m = p[i];
            toret = i;
        }
    }
    return toret;
}
#endif

/* v in -1.0 to 1.0 */
double quantize( double v, int steps ) {
    steps--;
    v = v + 1.0;
    v = v / 2.0;
    v = v * steps;
#ifdef __linux__
    v = rint(v);
#else
    v = round(v);
#endif
    v = v / ((double)steps);
    v = v * 2.0;
    v = v - 1.0;
    return v;
}

double VotingSystem::pickOneHappiness( const VoterArray& they, int numv, int winner, double* stddevP, int start ) {
    double happiness = 0;
    int i, end;
    end = start + numv;
    for ( i = start; i < end; i++ ) {
        happiness += they[i].getPref( winner );
    }
    happiness = happiness / numv;
    if ( stddevP != NULL ) {
	double stddev = 0;
	for ( i = start; i < end; i++ ) {
	    double d;
	    d = they[i].getPref( winner ) - happiness;
	    stddev += ( d * d );
	}
	stddev = stddev / numv;
	stddev = sqrt( stddev );
	*stddevP = stddev;
    }
    return happiness;
}
double VotingSystem::pickOneHappiness( const VoterArray& they, int numv, int winner, double* stddevP ) {
    return pickOneHappiness( they, numv, winner, stddevP, 0 );
}
double VotingSystem::pickOneHappiness( const VoterArray& they, int numv, int winner ) {
    double happiness = 0;
    int i;
    for ( i = 0; i < numv; i++ ) {
        happiness += they[i].getPref( winner );
    }
    happiness = happiness / numv;
    return happiness;
}

void voterDump( char* voteDumpFilename, const VoterArray& they, int numv, int numc ) {
    FILE* voteDumpFile;
    voteDumpFile = fopen( voteDumpFilename, "w" );
	voterDump( voteDumpFile, they, numv, numc );
    fclose( voteDumpFile );
}
void voterDump( FILE* voteDumpFile, const VoterArray& they, int numv, int numc ) {
    fprintf( voteDumpFile, "%d\n%d\n", numv, numc );
    for ( int v = 0; v < numv; v++ ) {
	they[v].dump( voteDumpFile );
    }
	fflush( voteDumpFile );
}
void voterBinDump( char* voteBinDumpFilename, const VoterArray& they, int numv, int numc ) {
    FILE* voteDumpFile;
    voteDumpFile = fopen( voteBinDumpFilename, "w" );
    fwrite( &numv, sizeof(numv), 1, voteDumpFile );
    fwrite( &(numc), sizeof(numc), 1, voteDumpFile );
    for ( int v = 0; v < numv; v++ ) {
	they[v].write( voteDumpFile );
    }
    fclose( voteDumpFile );
}

// inline function to avoid double evaluation of a or b
inline double fmax( double a, double b ) {
    return (a>b)?a:b;
}

class MaxHappiness : public VotingSystem {
public:
MaxHappiness() : VotingSystem("Max Happiness") {};
virtual void runElection( int* winnerR, const VoterArray& they ) {
    double maxh;
    int c = 0;
    int numc = they.numc;
    int numv = they.numv;
    
    maxh = pickOneHappiness( they, numv, c );
    for ( int i = 1; i < numc; i++ ) {
        double h;
        h = pickOneHappiness( they, numv, i );
        if ( h > maxh ) {
            c = i;
            maxh = h;
        }
    }
    if ( winnerR ) *winnerR = c;
};
};

volatile int goGently = 0;

void mysigint( int a ) {
    goGently = 1;
}

class workQThread {
public:
    VoterSim* sim;
    DBResultFile* drf;
    NameBlock* nb;
    WorkSource* q;
    pthread_t thread;
    
    workQThread( VoterSim* s, DBResultFile* d, NameBlock* n, WorkSource* q );
};
void* workQThreadProc( void* a ) {
    workQThread* me = (workQThread*)a;
    me->sim->runFromWorkQueue( me->drf, *(me->nb), me->q );
    return NULL;
}
workQThread::workQThread( VoterSim* s, DBResultFile* d, NameBlock* n, WorkSource* qIn )
    : sim( s ), drf( d ), nb( n ), q( qIn )
{
    pthread_create( &thread, NULL, workQThreadProc, this );
}

int main( int argc, char** argv ) {
VotingSystem* systems[] = {
    new MaxHappiness(),
    new OneVotePickOne(),
    new InstantRunoffVotePickOne(),
    new AcceptanceVotePickOne(),
    new FuzzyVotePickOne( "Rated Vote, raw", 0, 0 ),
    new FuzzyVotePickOne( "Rated Vote, equal sum", 1, 0 ),
    new FuzzyVotePickOne( "Rated Vote, maximized", 2, 0 ),
    new FuzzyVotePickOne( "Rated Vote, 1..num choinces", 0, -1 ),
    new FuzzyVotePickOne( "Rated Vote, 1..10", 0, 10 ),
    new RankedVotePickOne(),
    new RankedVotePickOne("Ranked Vote, no neg pref",1),
    new Condorcet(),
    new IRNR("IRNR"),
    new RandomElection("Random"),
    NULL
};
    Steps* stepq = NULL;
    VoterSim* s;
    workQThread** wqts;
    DBResultFile* drf = NULL;
    char* drfname = NULL;
    NameBlock nb;
    int j = 0, i;
    int nsys;
    int numThreads = 1;
    int hang = 0;

    srandom(time(NULL));

    nsys = 0;
    {VotingSystem** curSys = systems; while ( *curSys ) {
        curSys++;
        nsys++;
    }}
    votingSystemArrayToNameBlock( &nb, systems, nsys );

    for ( i = 0; i < argc; i++ ) {
	if ( j != i ) {
	    argv[j] = argv[i];
	}
	if ( ! strcmp( argv[i], "-F" ) ) {
	    i++;
	    drfname = argv[i];
	} else if ( ! strcmp( argv[i], "--vsteps" ) ) {
	    i++;
	    if ( stepq == NULL ) stepq = new Steps;
	    stepq->parseV( argv[i] );
	} else if ( ! strcmp( argv[i], "--csteps" ) ) {
	    i++;
	    if ( stepq == NULL ) stepq = new Steps;
	    stepq->parseC( argv[i] );
	} else if ( ! strcmp( argv[i], "--esteps" ) ) {
	    i++;
	    if ( stepq == NULL ) stepq = new Steps;
	    stepq->parseE( argv[i] );
	} else if ( (! strcmp( argv[i], "-n" )) && (stepq != NULL) ) {
	    i++;j++;
	    argv[j] = argv[i];
	    stepq->n = atoi( argv[i] );
	    j++;
	} else if ( ! strcmp( argv[i], "--threads" ) ) {
	    i++;
	    numThreads = atoi( argv[i] );
	} else if ( ! strcmp( argv[i], "--hang" ) ) {
	    hang = 1;
	} else {
	    j++;
	}
    }
    argv[j] = argv[i];
    argc = j;

    if ( drfname != NULL ) {
	printf("opening result db \"%s\"...", drfname );
	if ( numThreads == 1 ) {
	    drf = DBResultFile::open( drfname );
	} else {
	    drf = ThreadSafeDBRF::open( drfname );
	}
	if ( drf ) {
	    printf("success\n");
	    drf->useNames( &nb );
	} else {
	    printf("failed\n");
	}
    }
    
    signal( SIGINT, mysigint );
    
    s = new VoterSim[numThreads];
    for ( int i = 0; i < numThreads; i++ ) {
	s[i].systems = systems;
	s[i].nsys = nsys;
#if 0
	for ( int ai = 0; ai < argc; ai++ ) {
	    printf("argv[%d] \"%s\"\n", ai, argv[ai] );
	}
#endif
	s[i].init( argc, argv );
    }
    if ( stepq == NULL ) {
	s[0].run( drf, nb );
    } else if ( numThreads == 1 ) {
	printf("running work set\n");
	s[0].runFromWorkQueue( drf, nb, stepq );
    } else {
	WorkSource* q;
	printf("running %d threads...", numThreads);
	q = new ThreadSafeWorkSource( stepq );
	wqts = new workQThread*[numThreads];
	for ( int i = 0; i < numThreads; i++ ) {
	    wqts[i] = new workQThread( s + i, drf, &nb, q );
	}
	printf("running\n");
	for ( int i = 0; i < numThreads; i++ ) {
	    pthread_join( wqts[i]->thread, NULL );
	    printf("thread %d done\n", i );
	}
	printf("all threads done\n");
	for ( int i = 0; i < numThreads; i++ ) {
	    delete wqts[i];
	}
	delete [] wqts;
	delete q;
    }
    if ( hang ) {
	goGently = 0;
	while ( ! goGently ) {}
    }
    if ( drf ) {
	drf->flush();
	drf->close();
    }
    delete [] s;
    if ( hang ) {
	goGently = 0;
	while ( ! goGently ) {}
    }
    return 0;
}
